import pathlib
import pivtestdata as ptd
import unittest

import pivimage

__this_dir__ = pathlib.Path(__file__).parent


class TestMetadata(unittest.TestCase):

    def setUp(self) -> None:
        pc_1B = ptd.piv_challenge.pc_1B
        sample_folder_pc_1B = pc_1B.download()
        self.filenames = sorted(sample_folder_pc_1B.glob('*.tif*'))

    def test_dict(self):
        pivimg = pivimage.PIVImage(self.filenames[0],
                                   meta={'name': 'test image'})
        self.assertIsInstance(pivimg.meta, pivimage.meta.Metadata)
        pivimg.meta.save(filename='test.json')
        filename = pivimg.meta.save()
        self.assertTrue(filename.exists())
        filename.unlink(missing_ok=True)

        no_filename_pivimg = pivimage.PIVImage(None)
        with self.assertRaises(ValueError):
            filename = no_filename_pivimg.meta.save()



import numpy as np
import pathlib
import pivtestdata as ptd
import unittest

import pivimage

__this_dir__ = pathlib.Path(__file__).parent


class TestPivImages(unittest.TestCase):

    def setUp(self) -> None:
        pc_1B = ptd.piv_challenge.pc_1B
        sample_folder_pc_1B = pc_1B.download()
        self.filenames = sorted(sample_folder_pc_1B.glob('*.tif*'))

    def test_PIVImages(self):
        pivimgs = pivimage.PIVImages(self.filenames)
        self.assertEqual(len(pivimgs), len(self.filenames))
        self.assertEqual(pivimgs[0].filename, self.filenames[0])

    def test_background(self):
        pivimgs = pivimage.PIVImages(self.filenames)
        bg = pivimgs.compute_background(np.min)
        bg.rot180()



import abc
import pathlib
from typing import Dict, Union

from .file_wrapper import FileWrapper


def _readonly(self, *args, **kwargs):
    raise RuntimeError("Cannot modify ReadOnlyDict")


class ReadOnlyDict(dict):
    """from https://stackoverflow.com/questions/19022868/how-to-make-dictionary-read-only"""
    __setitem__ = _readonly
    __delitem__ = _readonly
    pop = _readonly
    popitem = _readonly
    clear = _readonly
    update = _readonly
    setdefault = _readonly


class MetaFileInterface(abc.ABC):
    """Abstract base class for metadata interfacing"""

    def __init__(self, hidden: bool = False):
        self.hidden = hidden

    @classmethod
    @abc.abstractmethod
    def load(cls, filename: Union[str, pathlib.Path]) -> Dict:
        """load a metadata file"""

    @classmethod
    @abc.abstractmethod
    def save(cls, data, filename: Union[str, pathlib.Path]) -> pathlib.Path:
        """save a metadata file"""


class Metadata:

    def __init__(self,
                 parent: FileWrapper):
        self._parent_filename = parent.filename
        self._data = {}
        self._filename = None

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.filename}")'

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self._data[k] = v

    def add(self, key, value):
        self.update({key: value})

    def load(self,
             interface: MetaFileInterface = None,
             readonly=True) -> ReadOnlyDict:
        """Load metadata from file even if already loaded."""
        if interface is None:
            interface = metafile_interfaces[self.filename.suffix[1:]]
        data = interface.load(self.filename)
        self._data = ReadOnlyDict(data) if readonly else data

        return self._data

    def save(self,
             filename: Union[str, pathlib.Path] = None,
             interface: MetaFileInterface = None):
        """Write metadata to file. The metadata interface class is determined
        based on the extension, e.g. '.json' will use `JSONMeta`"""
        if filename is None:
            filename = self.filename
        filename = pathlib.Path(filename)

        if interface is None:
            try:
                interface = metafile_interfaces[filename.suffix[1:]]
            except KeyError as e:
                av_interfaces = metafile_interfaces
                av_interfaces.pop('dict')
                raise KeyError(f'A Metadata interface class for {filename.suffix} does not exist. '
                               f'Please choose one of {", ".join(av_interfaces.keys())}')
        return interface.save(self._data, filename)

    def unlink(self, missing_ok: bool = True) -> None:
        """Deletes the meta file. cannot be recovered."""
        if self.filename is not None:
            self.filename.unlink(missing_ok=missing_ok)

    @property
    def filename(self) -> str:
        """build filename for metadata file"""
        if self._filename is None:
            # the default meta file ext is json
            if self._parent_filename is None:
                raise ValueError('The parent class is not associated to a file!')
            self._filename = self._parent_filename.with_suffix('.json')
        return self._filename

    @property
    def data(self) -> ReadOnlyDict:
        """Return metadata dict. loads if not already loaded."""
        if self._data is None:
            return self.load()
        return ReadOnlyDict(self._data)


class DictMeta(MetaFileInterface):
    """Meta data manager for data stored in python dictionary"""

    SUFFIX = None

    def load(self, filename):
        raise NotImplementedError(f'{self.__class__.__name__} has no load method implemented')

    def save(self, data=None, filename=None):
        raise NotImplementedError(f'{self.__class__.__name__} has no generic save method.')

    @classmethod
    def json_save(cls, data, filename):
        """save as json"""
        import json
        with open(filename) as f:
            return json.dump(data, f)

    @classmethod
    def json_yaml(cls, data, filename):
        """save as yaml"""
        import yaml
        with open(filename) as f:
            return yaml.safe_dump(data, f)


class JSONMeta(MetaFileInterface):
    """Meta data manager for data stored in JSON files"""

    SUFFIX = '.json'

    @classmethod
    def load(cls, filename):
        """load metadata"""
        import json
        with open(filename) as f:
            return json.load(f)

    @classmethod
    def save(cls, data, filename) -> pathlib.Path:
        """save metadata"""
        import json
        with open(filename, 'w') as f:
            json.dump(data, f)
        return pathlib.Path(filename)


class YAMLMeta(MetaFileInterface):
    """Meta data manager for data stored in YAML files"""

    SUFFIX = '.yaml'

    @classmethod
    def load(self, filename):
        """load metadata"""
        import yaml
        with open(filename) as f:
            return yaml.safe_load(f)

    @classmethod
    def save(self, data, filename):
        """save metadata"""
        import yaml
        with open(filename) as f:
            return yaml.safe_dump(data, f)


metafile_interfaces = {
    'json': JSONMeta,
    'yaml': YAMLMeta,
    'dict': DictMeta}


def metafile(reader: str, hidden: bool = True) -> MetaFileInterface:
    """automatically selects meta file reader"""
    return metafile_interfaces[reader](hidden=hidden)

# class Metadata:
#     def __init__(self, hidden: bool, reader: Union[str, MetaFileManager]):
#         self.hidden = hidden  # whether filenames are hidden, thus starting with a "."
#         self.reader = reader if isinstance(reader) else metadata_managers.get(reader, None)
#         if self.reader is None:
#             raise KeyError(f'Invalid reader: {reader}')
#         self._parent = None
#         self.filename = None
#
#     def assign(self, parent: FileWrapper):
#         pf = parent.filename
#         if self.hidden:
#             self.filename = pf.with_name(f'.{pf.stem}')
#         self.filename = pf.with_suffix(self.reader.SUFFIX)
#
#     def load(self):
#         if self.filename is None:
#             raise RuntimeError('Please first assign the Metadata object to a PIVImage using ".assign(<img>)"')
#         self._data = self.reader.load(self.filename)



import pathlib


class FileWrapper:
    """A (very basic) class associated with a file"""

    def __init__(self, filename, missing_ok=True):
        if filename is None:
            if not missing_ok:
                raise ValueError('filename cannot be None')
            self._filename = None
        else:
            self._filename = pathlib.Path(filename)
            if not missing_ok and not self._filename.exists():
                raise FileNotFoundError(f'File "{self._filename}" does not exist.')

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._filename}")'

    @property
    def filename(self):
        """Return filename"""
        return self._filename



import abc
import copy
import cv2
import logging
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import warnings
from cv2 import imread as cv2_imread
from pathlib import Path
from pco_tools import pco_reader
from typing import Union, List, Tuple, Dict

from .file_wrapper import FileWrapper
from .meta import Metadata

keep_meta = False
logger = logging.getLogger(__package__)


class _PIVImage(abc.ABC, FileWrapper):

    def __init__(self,
                 filename,
                 meta: Union[Dict, Metadata] = None,
                 **kwargs):
        super().__init__(filename, **kwargs)
        self._img = None

        attrs = kwargs.get('attrs', None)
        if attrs is not None:
            warnings.warn(f'Parameter "attrs" is deprecated. Please use "meta" instead.')
            if meta is not None:
                raise ValueError(f'Both "meta" and "attrs" are given. In future only "meta" will be available. '
                                 f'Make sure to only pass one of it at this stage, preferably "meta".')

        if isinstance(meta, dict):
            self._meta = Metadata(self)
        else:
            self._meta = Metadata(self)

    @property
    def meta(self):
        """Return metadata object"""
        return self._meta

    @meta.setter
    def meta(self, data: Dict):
        if not isinstance(data, dict):
            raise TypeError('Only allow a dictionary')
        self._meta._data = data

    def __getitem__(self, item):
        return self.get().__getitem__(item)

    def asxarray(self):
        """Use xarray instead of numpy array. Requires xarray to be installed.
        A new object with xarray.DataArray for the iamge array is returned.
        """
        try:
            import xarray as xr
        except ImportError:
            raise ImportError('xarray is not installed. Please install it first.')
        ny, nx = self.shape
        new_obj = copy.deepcopy(self)
        try:
            attrs = self.meta.data
        except NotImplementedError as e:
            attrs = {}
        new_obj._img = xr.DataArray(new_obj.get(),
                                    dims=('iy', 'ix'),
                                    coords={'iy': np.arange(ny),
                                            'ix': np.arange(nx)},
                                    attrs=attrs)
        return new_obj

    def unlink(self, missing_ok: bool = True):
        """Deletes the image file and the meta file (if exists/associated)"""
        self.filename.unlink(missing_ok=missing_ok)
        self.meta.unlink(missing_ok=missing_ok)

    @property
    def filename(self):
        """Return filename"""
        return self._filename

    def __array__(self) -> np.ndarray:
        """returns numpy array if np.asarray() or e.g. np.roll() is called on this object"""
        return np.asarray(self.get())

    @classmethod
    @abc.abstractmethod
    def from_array(cls, img: np.ndarray, meta: Union[Dict, Metadata] = None):
        """Initialize from array"""

    @property
    def ndim(self):
        """Return ndim of array"""
        return self.get().ndim

    @property
    def dtype(self):
        return self.get().dtype

    @property
    def shape(self):
        """Return shape of array"""
        return self.get().shape

    def clear(self):
        """free the (RAM), aka unset _img"""
        self._img = None

    def get(self) -> np.ndarray:
        """Return the image. If not yet loaded (self._img is None), load it."""
        if self._img is None:
            if self._filename is None:
                raise ValueError('No filename set!')
            self._img = loadimg(self._filename)
        return self._img

    def max(self):
        """Return maximum value of the image"""
        return self.get().max()

    def min(self):
        """Return minimum value of the image"""
        return self.get().min()

    def mean(self, *args, **kwargs):
        """Return the mean value of the image"""
        return self.get().mean(*args, **kwargs)

    def median(self, *args, **kwargs):
        """Return the median value of the image"""
        return np.median(self.get(), *args, **kwargs)

    def std(self, *args, **kwargs):
        """Return the standard deviation value of the image"""
        return np.std(self.get(), *args, **kwargs)

    def grayscale(self) -> "PIVImage":
        """make image grayscale if ndim==3"""
        if self.ndim == 2:
            return copy.deepcopy(self)
        new_piv_image = PIVImage.from_array(self._img.copy(), is_first_image=self._is_a, pco=self.pco)
        new_piv_image._img = cv2.cvtColor(self.get(), cv2.COLOR_BGR2GRAY)
        return new_piv_image

    def smooth(self, kernel_size) -> "PIVImage":
        """smooth the image"""
        kernel = np.ones((kernel_size, kernel_size), np.float32) / kernel_size ** 2
        smoothed_image = cv2.filter2D(self.get(), -1, kernel)
        return PIVImage.from_array(smoothed_image, meta=self.meta.copy() if keep_meta else {})

    def normalize(self, inplace: bool = False) -> "PIVImage":
        """normalize the image"""
        _img = self.get()
        _min = np.nanmin(_img)
        _max = np.nanmax(_img)
        norm_img = (_img - _min) / (_max - _min)
        if inplace:
            self._img = norm_img
            return self
        return self.from_array(norm_img, meta=self.meta.copy() if keep_meta else {})

    def rot90(self, inplace: bool = False) -> "PIVImage":
        """rotate image by 90 degrees"""
        _img = self.get()
        if inplace:
            self._img = np.rot90(_img)
            return self
        return self.from_array(np.rot90(_img), meta=self.meta.copy() if keep_meta else {})

    def fliplr(self, inplace: bool = False) -> "PIVImage":
        """flip image left-right"""
        _img = self.get()
        if inplace:
            self._img = np.fliplr(_img)
            return self
        return self.from_array(np.fliplr(_img), meta=self.meta.copy() if keep_meta else {})

    def flipud(self, inplace: bool = False) -> "PIVImage":
        """flip image up-down"""
        _img = self.get()
        if inplace:
            self._img = np.flipud(_img)
            return self
        return self.from_array(np.flipud(_img), meta=self.meta.copy() if keep_meta else {})

    def rot180(self, inplace: bool = False) -> "PIVImage":
        """rotate image by 180 degrees"""
        _img = self.get()
        if inplace:
            self._img = np.rot90(_img, k=2)
            return self
        return self.from_array(np.rot90(_img, k=2), meta=self.meta.copy() if keep_meta else {})

    def apply_mask(self, mask_array: np.ndarray, fill_value: int, inplace: bool = False) -> "PIVImage":
        """Apply a mask to the image. Will set the masked values to `fill_value`."""
        masked_img = self.get().copy()
        masked_img[mask_array] = fill_value
        if inplace:
            self._img = masked_img
            return self
        return self.from_array(masked_img, meta=self.meta.copy() if keep_meta else {})

    def plot(self,
             figure_height: float = 3.,
             spacing: float = 0.0,
             ax_hist_ratio: float = 0.05,
             vmin: float = None,
             vmax: float = None,
             bins: int = 101,
             density: bool = False,
             axs: Tuple[plt.Axes, None] = None):
        """plot the image. If no `ax` is provided a histogram is automatically plotted below the image."""
        figure_height = figure_height
        spacing = spacing

        _shape = self.get().shape
        h, w = _shape[0], _shape[1]
        hist_height = ax_hist_ratio * figure_height

        left, width = 0.1, w / w
        bottom, height = 0.1, h / h

        hist_ax_pos = [left, bottom, width, hist_height]
        img_ax_pos = [left, hist_height + bottom + spacing, width, height]

        if axs is not None:
            if len(axs) == 1:
                return self._plot(ax=axs[0], cmap='gray', vmin=vmin, vmax=vmax, hide_colorbar=True)
            ax_img = axs[0]
            ax_hist = axs[1]
        else:
            # start with a square Figure
            fig = plt.figure(figsize=(w / h * figure_height, figure_height))

            ax_img = fig.add_axes(img_ax_pos)
            ax_img.axis('off')

            # if self._is_a is not None:
            #     if self._is_a:
            #         ax_img.set_title('A', size=12)
            #     else:
            #         ax_img.set_title('B', size=12)

            ax_hist = fig.add_axes(hist_ax_pos)

        self._plot(ax=ax_img, cmap='gray', vmin=vmin, vmax=vmax, hide_colorbar=True)
        self.hist(ax=ax_hist, bins=bins, color='k', density=density)

        if vmax:
            ax_hist.vlines(vmax, 0, ax_hist.get_ylim()[1], linestyles='--', color='gray')
        if vmin:
            ax_hist.vlines(vmin, 0, ax_hist.get_ylim()[1], linestyles='--', color='gray')

        return ax_img, ax_hist

    def _plot(self, ax=None, autoscale: bool = False, **kwargs):
        if ax is None:
            ax = plt.gca()
        _img = self.get()
        if autoscale:
            _vmin = kwargs.get('vmin', np.nanmin(_img))
            kwargs['vmin'] = _vmin
            _vmax = kwargs.get('vmax', np.nanmin(_img))
            kwargs['vmax'] = _vmax

        cmap = kwargs.pop('cmap', 'gray')
        hide_colorbar = kwargs.pop('hide_colorbar', False)
        im = ax.imshow(_img, cmap=cmap, **kwargs)
        if not hide_colorbar:
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(im, cax=cax)
        return ax

    def hist(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        color = kwargs.pop('color', 'k')
        ax.hist(self.get().ravel(), color=color, **kwargs)
        return ax

    def to_tiff(self, filename, write_meta: bool = False):
        filename = save_image(filename, self[:])
        if write_meta:
            yam_filename = filename.parent / f'{filename.stem}_meta.yaml'
            try:
                import yaml
            except ImportError:
                raise ImportError('Please install pyyaml to write the meta to a yaml file.')
            with open(yam_filename, 'w') as f:
                yaml.safe_dump(self.meta, f)
        return filename


class PIVBackgroundImage(_PIVImage):
    """PIV background image helper class"""

    @classmethod
    def from_array(cls, img: np.ndarray, meta: Union[Dict, Metadata] = None):
        """Create a PIVImage object from a numpy array"""
        obj = cls(filename=None, meta=meta)
        obj._img = img
        return obj


class PIVImage(_PIVImage):
    """PIV image helper class"""

    def __init__(self,
                 filename: Union[pathlib.Path, None],
                 is_first_image: bool = None,
                 pco: bool = False,
                 meta: Union[Dict, Metadata] = None,
                 **kwargs):
        super().__init__(filename=filename, meta=meta, **kwargs)

        self.pco = pco
        self._is_a = is_first_image
        if self._is_a is None and pco:
            raise ValueError('If pco is set, ou must specify whether this is the first or second image via '
                             'parameter "is_first_image".')
        self._img = None

    def __sub__(self, other):
        """Subtract two PIV images. If the result is negative, set it to zero."""
        if not isinstance(other, (PIVImage, np.ndarray)):
            raise TypeError(f'Cannot subtract {type(other)} from {type(self)}')
        if isinstance(other, PIVImage):
            diffimg = self.get() - other.get()
        else:
            diffimg = self.get() - other
        diffimg[diffimg < 0] = 0
        return self.__class__(filename=None, is_first_image=None).from_array(diffimg)

    def pair_with(self, other: _PIVImage) -> "PIVImagePair":
        """Pair this image with another and return an image pair"""
        return PIVImagePair(self, other)

    def get(self) -> np.ndarray:
        """Return the image. If not yet loaded (self._img is None), load it."""
        if self._img is None:
            if self._filename is None:
                raise ValueError('No filename set!')
            img = loadimg(self._filename)
            if not self.pco:
                self._img = img
            else:
                ny, nx = img.shape
                if self._is_a:
                    print('return second part of image')
                    self._img = img[:ny // 2, ...]
                else:
                    print('return first part of image')
                    self._img = img[ny // 2:, ...]
        return self._img

    @classmethod
    def from_array(cls,
                   arr,
                   is_first_image: bool = None,
                   pco: bool = False,
                   meta: Union[Dict, Metadata] = None,
                   **kwargs) -> "PIVImage":
        """Init the class from a numpy array."""
        pivimg = cls(filename=None, is_first_image=is_first_image, pco=pco, meta=meta, **kwargs)
        ny, nx = arr.shape[0], arr.shape[1]
        if pco and is_first_image:
            pivimg._img = arr[:ny // 2, ...]
        elif pco and not is_first_image:
            pivimg._img = arr[ny // 2:, ...]
        else:
            pivimg._img = arr
        return pivimg


class PIVImages:
    """Collection of PIV images. The images are loaded on demand. The images have no
    relation to each other. They are just a collection of images. If you intend to work
    with image pairs, use the PIVImagePair class instead."""

    def __init__(self, filenames: List):
        self.filenames = list(filenames)
        self._images = {}

    def __len__(self):
        return len(self.filenames)

    def __getitem__(self, item) -> PIVImage:
        """Return image array. Loaded on demand, which means, that if the image was
        loaded before, it will be returned from memory, otherwise it will be loaded from
        disk."""
        if isinstance(item, slice):
            return PIVImages.from_array_list([self[i] for i in range(*item.indices(len(self)))])
        if item not in self._images:
            self._images[item] = PIVImage(self.filenames[item])
        return self._images[item]

    def compute_background(self, func) -> PIVImage:
        """Compute the background image using the given function. The function must
        accept a numpy array and axis=0 as input and return a numpy array as output.
        E.g.: np.min, np.mean, np.median, etc.

        Parameters
        ----------
        func : Callable
            Function to compute the background image. Must accept a numpy array and
            axis=0 as input and return a numpy array as output.

        Returns
        -------
        PIVImage
            Background image.
        """
        return PIVBackgroundImage.from_array(func(np.stack([self[i][:] for i in range(len(self))]), axis=0))

    @classmethod
    def from_array_list(cls, list_of_arrays: List[np.ndarray]):
        """Create a PIVImages object from a list of numpy arrays."""
        obj = cls([None for _ in range(len(list_of_arrays))])
        obj._images = {i: PIVImage(list_of_arrays[i]) for i in range(len(list_of_arrays))}
        return obj


class PIVImagePair:
    """Helper class to work with a pair of PIV images

    If a PIV recording is only stored in one file, than pass None for filename_B.
    """

    def __init__(self,
                 A: Union[str, pathlib.Path, PIVImage],
                 B: Union[str, pathlib.Path, PIVImage]):
        if A is None:
            raise ValueError('A cannot be None!')

        if isinstance(A, PIVImage):
            self._A = A
        else:
            self._A = PIVImage(A, is_first_image=True)

        if isinstance(B, PIVImage):
            self._B = B
        else:
            self._B = PIVImage(B, is_first_image=False)

    def plot(self,
             figure_height: float = 3.,
             spacing: float = 0.0,
             ax_hist_ratio: float = 0.05,
             vmin: float = None,
             vmax: float = None,
             bins: int = 101,
             density: bool = False):
        """Plot both images next to each other"""
        figure_height = figure_height
        spacing = spacing

        _shape = self.A.get().shape
        h, w = _shape[0], _shape[1]
        hist_height = ax_hist_ratio * figure_height

        left, width = 0.1, w / w
        bottom, height = 0.1, h / h

        # start with a square Figure
        fig = plt.figure(figsize=(w / h * figure_height, figure_height))

        hist_ax_pos_A = [left, bottom, width, hist_height]
        img_ax_pos_A = [left, hist_height + bottom + spacing, width, height]
        ax_imgA = fig.add_axes(img_ax_pos_A)
        ax_imgA.axis('off')
        ax_histA = fig.add_axes(hist_ax_pos_A)

        eps = 0.001

        hist_ax_pos_B = [left + hist_ax_pos_A[2]+eps, bottom, width, hist_height]
        img_ax_pos_B = [left + img_ax_pos_A[2]+eps, hist_height + bottom + spacing, width, height]
        ax_imgB = fig.add_axes(img_ax_pos_B)
        ax_imgB.axis('off')
        ax_histB = fig.add_axes(hist_ax_pos_B)

        self.A._plot(ax=ax_imgA, cmap='gray', vmin=vmin, vmax=vmax, hide_colorbar=True)
        self.A.hist(ax=ax_histA, bins=bins, color='k', density=density)

        self.B._plot(ax=ax_imgB, cmap='gray', vmin=vmin, vmax=vmax, hide_colorbar=True)
        self.B.hist(ax=ax_histB, bins=bins, color='k', density=density)

        if vmax:
            ax_histA.vlines(vmax, 0, ax_histA.get_ylim()[1], linestyles='--', color='gray')
        if vmin:
            ax_histA.vlines(vmin, 0, ax_histA.get_ylim()[1], linestyles='--', color='gray')

        if vmax:
            ax_histB.vlines(vmax, 0, ax_histB.get_ylim()[1], linestyles='--', color='gray')
        if vmin:
            ax_histB.vlines(vmin, 0, ax_histB.get_ylim()[1], linestyles='--', color='gray')

        ax_hist_A_ymax = ax_histA.get_ylim()[1]
        ax_hist_B_ymax = ax_histB.get_ylim()[1]
        common_hist_ymax = max(ax_hist_A_ymax, ax_hist_B_ymax)
        ax_histA.set_ylim([0, common_hist_ymax])
        ax_histB.set_ylim([0, common_hist_ymax])

        # disable ax_histB yticks
        ax_histB.set_yticks([])

        return ax_imgA, ax_histA, ax_imgB, ax_histB

    def plot_overlay(self, channel_A=0, channel_B=2, **kwargs):
        """plot both images in different channels. A is plotted in red, B in blue"""
        if channel_A not in [0, 1, 2]:
            raise ValueError('channel_A must be 0, 1 or 2')
        if channel_B not in [0, 1, 2]:
            raise ValueError('channel_B must be 0, 1 or 2')
        if channel_A == channel_B:
            raise ValueError('channel_A and channel_B must be different')
        ax = kwargs.pop('ax', plt.gca())
        arrA = self.A.get()
        arrRGB = np.zeros((*arrA.shape, 3))
        arrRGB[:, :, channel_A] = arrA
        arrRGB[:, :, channel_B] = self.B.get()
        ax.imshow(arrRGB)
        return ax

    def subtract_background(self, background_filename: Union[Union[str, pathlib.Path],
                                                             List[Union[str, pathlib.Path]]]) -> "PIVImagePair":
        """subtract the background from both images. A new PIVImagePair is returned"""
        if not isinstance(background_filename, (tuple, list)):
            background_filename = [background_filename, ]
        if len(background_filename) == 0:
            raise ValueError('No background file provided!')
        elif len(background_filename) > 2:
            raise ValueError('Too many background files provided!')
        bga = loadimg(background_filename[0])
        if len(background_filename) == 2:
            bgb = loadimg(background_filename[1])
        else:
            bgb = bga
        _piv_img_pari = copy.deepcopy(self)
        _piv_img_pari._A._img = _piv_img_pari._A._img - bga
        _piv_img_pari._B._img = _piv_img_pari._B._img - bgb
        return _piv_img_pari

    @property
    def A(self):
        return self._A

    @property
    def B(self):
        return self._B


class PIVImagePairs:
    """Collection of PIV images paris"""

    def __init__(self, filenames_A: list, filenames_B: list):
        self.filenames_A = filenames_A
        self.filenames_B = filenames_B
        self._images_A = {}
        self._images_B = {}

    def __getitem__(self, item):
        """Return image array"""
        if item not in self._images_A:
            self._images_A[item] = PIVImage(self.filenames_A[item])
        if item not in self._images_B:
            self._images_B[item] = PIVImage(self.filenames_B[item])
        return PIVImagePair(self._images_A[item], self._images_B[item])


def load_piv_image(filename: Path, is_first: bool = None):
    """Initialize a PIVImage from a filename"""
    return PIVImage(filename, is_first)


def loadimg(img_filepath: Path):
    """
    loads b16 or other file format
    """
    img_filepath = Path(img_filepath)
    if img_filepath.suffix in ('b16', '.b16'):
        im_ = pco_reader.load(str(img_filepath))
    else:
        im_ = cv2_imread(str(img_filepath), -1)
    return im_


def save_image(filename, arr, img_type='tiff'):
    """write data to a tiff file"""
    if img_type in ('tiff', '.tiff'):
        cv2.imwrite(str(filename), arr)
        return pathlib.Path(filename).absolute()
    raise NotImplementedError('Image types other than tiff are not implemented yet.')



try:
    from importlib.metadata import version as _version
except ImportError as e:
    raise ImportError('Most likely you have python<3.8 installed or in case you '
                      'ran `pytest` you may not have installed it for the correct '
                      'python version. At least 3.8 is required.')

__version__ = _version('pivimage')



import cv2
import numpy as np
import pathlib
import pivtestdata as ptd
import unittest
import xarray as xr

import pivimage

__this_dir__ = pathlib.Path(__file__).parent


class TestPivImage(unittest.TestCase):

    def setUp(self) -> None:
        pc_1B = ptd.piv_challenge.pc_1B
        sample_folder_pc_1B = pc_1B.download()
        self.filenames = sorted(sample_folder_pc_1B.glob('*.tif*'))

    def test_array(self):
        pivimg = pivimage.PIVImage(self.filenames[0])
        self.assertEqual(pivimg._img, None)
        self.assertIsInstance(pivimg.__array__(), np.ndarray)

        pivimg = pivimage.PIVImage(self.filenames[0])
        self.assertIsInstance(np.asarray(pivimg), np.ndarray)

        xrpivimg = pivimg.asxarray()
        self.assertIsInstance(xrpivimg[:], xr.DataArray)

    def test_PIVImage(self):
        pivimg_none = pivimage.PIVImage(None)
        self.assertEqual(pivimg_none._filename, None)
        with self.assertRaises(ValueError):
            pivimg_none.get()

        with self.assertRaises(ValueError):
            _ = pivimage.PIVImage(self.filenames[0], pco=True)
        pivimg = pivimage.PIVImage(self.filenames[0],
                                   pco=False,
                                   is_first_image=True)
        self.assertEqual(pivimg._img, None)
        self.assertIsInstance(pivimg._filename, pathlib.Path)
        self.assertEqual(pivimg._filename, self.filenames[0])
        self.assertEqual(pivimg.filename, self.filenames[0])
        ref_img = pivimage.loadimg(self.filenames[0])
        self.assertFalse(pivimg.pco)
        self.assertEqual(pivimg[:].shape, ref_img.shape)
        np.testing.assert_array_equal(pivimg[:], ref_img)
        np.testing.assert_array_equal(pivimg[0:10, 0:4:2], ref_img[0:10, 0:4:2])
        self.assertEqual(pivimg.shape, ref_img.shape)
        self.assertEqual(pivimg.ndim, ref_img.ndim)
        pivimg.clear()
        self.assertEqual(pivimg._img, None)
        self.assertEqual(pivimg.ndim, ref_img.ndim)
        pivimg.clear()
        self.assertEqual(pivimg.shape, ref_img.shape)
        pivimg.clear()
        np.testing.assert_array_equal(pivimg[0:10, 0:4:2], ref_img[0:10, 0:4:2])

        pivimg = pivimage.PIVImage.from_array(ref_img)
        self.assertEqual(pivimg._filename, None)
        self.assertEqual(pivimg.filename, None)
        self.assertEqual(pivimg[:].shape, ref_img.shape)

        with self.assertRaises(ValueError):
            _ = pivimage.PIVImage.from_array(ref_img, pco=True)
        pivimg = pivimage.PIVImage.from_array(ref_img)
        self.assertFalse(pivimg._img is None)
        np.testing.assert_array_equal(pivimg[:], ref_img[:])

        pivimg = pivimage.PIVImage.from_array(ref_img,
                                              pco=True,
                                              is_first_image=True)
        ny, nx = ref_img.shape
        np.testing.assert_array_equal(pivimg[:], ref_img[:ny // 2, :])

        pivimg = pivimage.PIVImage.from_array(ref_img,
                                              pco=True,
                                              is_first_image=False)
        ny, nx = ref_img.shape
        np.testing.assert_array_equal(pivimg[:], ref_img[ny // 2:, :])

    def test_max_min(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        self.assertEqual(pivimg.max(), ref_img.max())
        self.assertEqual(pivimg.min(), ref_img.min())

    def test_mask(self):
        for fill_value in [0, 50.4, 100]:
            pivimg = pivimage.PIVImage(self.filenames[0])
            mask = np.zeros_like(pivimg[:], dtype=bool)
            mask[0:100, 0:40] = True
            masked_pivimg = pivimg.apply_mask(mask, fill_value)
            with self.assertRaises(AssertionError):
                np.testing.assert_array_equal(masked_pivimg[:], pivimg[:])
            np.testing.assert_array_equal(masked_pivimg[0:100, 0:40], int(fill_value) * np.ones((100, 40)))

            masked_pivimg2 = pivimg.apply_mask(mask, fill_value, inplace=True)
            self.assertTrue(masked_pivimg2 is pivimg)

    def test_sub(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        with self.assertRaises(TypeError):
            _ = pivimg - 1
        new_pivimg = pivimg - ref_img
        np.testing.assert_array_equal(new_pivimg._img, np.zeros_like(ref_img))

    def test_arraywrapper(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        self.assertIsInstance(np.asarray(pivimg), np.ndarray)

    def test_grayscale(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg_gray = pivimg.grayscale()
        self.assertFalse(pivimg is pivimg_gray)

        rgb = np.random.randint(255, size=(100, 50, 3), dtype=np.uint8)
        pivimg = pivimage.PIVImage.from_array(rgb)
        pivimg_gray = pivimg.grayscale()
        self.assertFalse(pivimg is pivimg_gray)
        self.assertEqual(3, rgb.ndim)
        self.assertEqual(3, pivimg.ndim)
        self.assertEqual(2, pivimg_gray.ndim)
        self.assertEqual((100, 50, 3), pivimg.shape)
        self.assertEqual((100, 50), pivimg_gray.shape)

        pivimg = pivimage.PIVImage.from_array(rgb, pco=True, is_first_image=True)
        pivimg_gray = pivimg.grayscale()
        self.assertFalse(pivimg is pivimg_gray)
        self.assertEqual(3, rgb.ndim)
        self.assertEqual(3, pivimg.ndim)
        self.assertEqual(2, pivimg_gray.ndim)
        self.assertEqual((50, 50, 3), pivimg.shape)
        self.assertEqual((50, 50), pivimg_gray.shape)

    def test_smooth(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg_smooth = pivimg.smooth(kernel_size=5)
        self.assertFalse(pivimg is pivimg_smooth)
        with self.assertRaises(cv2.error):
            pivimg.smooth(kernel_size=0)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg[:], pivimg_smooth[:])

    def test_normalize(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        self.assertTrue(pivimg[:].max() > 1)
        pivimg_norm = pivimg.normalize()
        self.assertFalse(pivimg is pivimg_norm)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg_norm[:], pivimg[:])
        self.assertEqual(pivimg_norm[:].max(), 1.0)
        self.assertEqual(pivimg_norm[:].min(), 0.0)

        max_value = pivimg.max()
        assert max_value > 1
        pivimg.normalize(inplace=True)
        self.assertEqual(pivimg[:].max(), 1.0)

    def test_rot90(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg_rot = pivimg.rot90()
        self.assertFalse(pivimg is pivimg_rot)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg_rot[:], pivimg[:])
        np.testing.assert_array_equal(pivimg_rot[:], np.rot90(ref_img, 1))

        pivimg_rot2 = pivimg_rot.rot90(inplace=True)
        self.assertTrue(pivimg_rot2 is pivimg_rot)

    def test_rot180(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg_rot = pivimg.rot180()
        self.assertFalse(pivimg is pivimg_rot)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg_rot[:], pivimg[:])
        np.testing.assert_array_equal(pivimg_rot[:], np.rot90(ref_img, 2))
        pivimg_rot2 = pivimg_rot.rot180()
        np.testing.assert_array_equal(pivimg_rot2[:], pivimg[:])

        pivimg_rot2 = pivimg_rot.rot180(inplace=True)
        self.assertTrue(pivimg_rot2 is pivimg_rot)

    def test_flip(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg_fliplr = pivimg.fliplr()
        self.assertFalse(pivimg is pivimg_fliplr)
        self.assertEqual(pivimg.shape, pivimg_fliplr.shape)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg_fliplr[:], pivimg[:])
        np.testing.assert_array_equal(pivimg_fliplr[:], np.fliplr(ref_img))

        pivimg_flipud = pivimg.flipud()
        self.assertFalse(pivimg is pivimg_flipud)
        self.assertEqual(pivimg.shape, pivimg_flipud.shape)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(pivimg_flipud[:], pivimg[:])
        np.testing.assert_array_equal(pivimg_flipud[:], np.flipud(ref_img))

        pivimg_flip2 = pivimg.fliplr(inplace=True)
        self.assertTrue(pivimg_flip2 is pivimg)

    def test_to_tiff(self):
        ref_img = pivimage.loadimg(self.filenames[0])
        pivimg = pivimage.PIVImage.from_array(ref_img)
        pivimg.to_tiff('test.tif')

        pivimg2 = pivimage.loadimg('test.tif')
        np.testing.assert_array_equal(pivimg2[:], pivimg[:])
        pathlib.Path('test.tif').unlink()



import pathlib
import unittest

import pivimage

__this_dir__ = pathlib.Path(__file__).parent


class TestVersion(unittest.TestCase):

    def test_version(self):
        this_version = 'x.x.x'
        setupcfg_filename = __this_dir__ / '../setup.cfg'
        with open(setupcfg_filename, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'version' in line:
                    this_version = line.split(' = ')[-1].strip()
        self.assertEqual(pivimage.__version__, this_version)



from ._version import __version__

from .meta import metafile
from .pivimage import PIVImage, PIVImages, PIVImagePair, PIVImagePairs, loadimg, save_image

__all__ = ["PIVImage", "PIVImages", "PIVImagePair", "PIVImagePairs", "loadimg",
           "metafile", "save_image", "__version__"]



import pathlib
import unittest

import pivtestdata as ptd

import pivimage

__this_dir__ = pathlib.Path(__file__).parent


class TestPivImagePair(unittest.TestCase):

    def setUp(self) -> None:
        pc_1B = ptd.piv_challenge.pc_1B
        sample_folder_pc_1B = pc_1B.download()
        self.filenames = sorted(sample_folder_pc_1B.glob('*.tif*'))

    def test_PIVImagePair(self):
        pivpair = pivimage.PIVImagePair(self.filenames[0], self.filenames[1])
        self.assertTrue(pivpair.A._img is None)
        imgA = pivpair.A[:]
        self.assertFalse(pivpair.A._img is None)
        self.assertIsInstance(pivpair._A, pivimage.PIVImage)
        self.assertIsInstance(pivpair.A, pivimage.PIVImage)
        self.assertEqual(pivpair._A.filename, self.filenames[0])
        self.assertIsInstance(pivpair._B, pivimage.PIVImage)
        self.assertIsInstance(pivpair.B, pivimage.PIVImage)
        self.assertEqual(pivpair._B.filename, self.filenames[1])
